package emst;

import java.awt.CheckboxMenuItem;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Menu;
import java.awt.MenuBar;
import java.awt.MenuItem;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.IOException;
import java.util.List;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;
import javax.swing.JColorChooser;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.UIManager;

import org.apache.log4j.BasicConfigurator;

import com.sun.opengl.util.Animator;
import com.sun.opengl.util.Screenshot;

import emst.algorithm.Kruskal;
import emst.algorithm.RandomWalk;
import emst.algorithm.VerifyST;
import emst.ds.Edge;
import emst.ds.Graph;
import emst.graphics.Camera;
import emst.util.Ms3dReader;

public class EmstApplication extends Frame implements WindowListener, GLEventListener, KeyListener, MouseListener, MouseMotionListener  {

	private static final long serialVersionUID = 1L;

	public static void main(String...args) {		
		BasicConfigurator.configure();
		
		try {
	        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	    } catch (Exception ex) {           
	        ex.printStackTrace();
	    }
		
		new EmstApplication();
	}
	
	// Menus
	private MenuBar menuBar;
	
	private Menu fileMenu;
	private MenuItem loadGraphMenuItem;
	private MenuItem loadMs3DFileMenuItem;
	private MenuItem reloadMs3DGraphFileMenuItem;
	private MenuItem reloadGraphFileMenuItem;
	private MenuItem saveScreenShotMenuItem;
	private MenuItem exitMenuItem;
	
	private Menu emstMenu;
	private MenuItem findEmstMenuItem;
	private MenuItem validateStMenuItem;
	
	private Menu viewMenu;
	private Menu graphViewMenu;
	private Menu emstViewMenu;
	
	private MenuItem graphViewEdgeColorMenuItem;
	private MenuItem graphViewVertexColorMenuItem;
	private CheckboxMenuItem graphViewToggleGraphMenuItem;
	
	private MenuItem emstViewColorMenuItem;
	private CheckboxMenuItem emstViewToggleEmst;
	
	boolean saveScreenshot;
	
	// file chooser for loading files
	private JFileChooser loadFileChooser;
	private JFileChooser saveScreenshotChooser;
	
	// Graph and Tree viewing options
	private Color graphEdgeColor;
	private Color graphVertexColor;
	
	private Color emstEdgeColor;
	
	// OpenGL Display Components
	private GLU glu;
	private GLCanvas glCanvas;
	private Animator animator;
	private int mouseClickX, mouseClickY;
	
	
	// Our own camera
	private Camera camera;
	
	// Graph and Tree
	private Graph graph;
	private List<Edge> tree;
	
	public EmstApplication() {
		
		saveScreenshot = false;
		
		emstEdgeColor = Color.RED;
		graphVertexColor = Color.CYAN;
		graphEdgeColor = Color.WHITE;
		
		setTitle("EMST Application");
		
		graph = null;
		tree = null;
		
		loadFileChooser = new JFileChooser(new File("."));
		saveScreenshotChooser = new JFileChooser(new File("."));
		
		buildMenus();
		
		addWindowListener(this);
		
		glu = new GLU();
		glCanvas = new GLCanvas();
		glCanvas.addGLEventListener(this);
		
		animator = new Animator(glCanvas);
		
		addWindowListener(new WindowAdapter() {
			public void windowClosing(WindowEvent e) {
				new Thread(new Runnable(){

					@Override
					public void run() {
						animator.stop();
						System.exit(0);						
					}}).start();				
			}
		});
		
		glCanvas.addKeyListener(this);
		add(glCanvas);
		
		this.setSize(800, 800);
		this.setPreferredSize(new Dimension(800, 800));
		
		
		pack();
		setVisible(true);
		
		animator.start();
	}
	
	
	private void buildMenus() {
		menuBar = new MenuBar();
		
		fileMenu = new Menu("File");
		menuBar.add(fileMenu);
		
		loadGraphMenuItem = new MenuItem("Load Graph File...");
		fileMenu.add(loadGraphMenuItem);
		loadGraphMenuItem.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				
				if (loadFileChooser.showOpenDialog(EmstApplication.this) == JFileChooser.APPROVE_OPTION) {
					try {
						graph = Graph.loadFile(loadFileChooser.getSelectedFile());
						tree = null;
						resetSettings();
					} catch (IOException ex) {
						ex.printStackTrace();
					}
				}				
			}
		});		
		
		loadMs3DFileMenuItem = new MenuItem("Load MS3D File...");
		fileMenu.add(loadMs3DFileMenuItem);
		loadMs3DFileMenuItem.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				
				if (loadFileChooser.showOpenDialog(EmstApplication.this) == JFileChooser.APPROVE_OPTION) {
					try {
						graph = Ms3dReader.readFile(loadFileChooser.getSelectedFile());
						graph.cleanUpGraph();
						tree = null;
						resetSettings();
					} catch (IOException e1) {
						e1.printStackTrace();
					}
				}				
			}
		});
		
		fileMenu.addSeparator();
		
		MenuItem randomWalkGraph = new MenuItem("Random Walk");
		randomWalkGraph.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				graph = RandomWalk.randomWalk(30000);
				tree = null;
				resetSettings();				
			}
		});
		fileMenu.add(randomWalkGraph);
		
		fileMenu.addSeparator();
		
		reloadGraphFileMenuItem = new MenuItem("Reload Current Graph File");
		fileMenu.add(reloadGraphFileMenuItem);
		reloadGraphFileMenuItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					graph = Graph.loadFile(loadFileChooser.getSelectedFile());
					tree = null;
					resetSettings();
				} catch (IOException e1) {
					e1.printStackTrace();
				}
				
			}});
		
		reloadMs3DGraphFileMenuItem = new MenuItem("Reload Current MS3D File");
		fileMenu.add(reloadMs3DGraphFileMenuItem);
		reloadMs3DGraphFileMenuItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				try {
					graph = Ms3dReader.readFile(loadFileChooser.getSelectedFile());
					tree = null;
					resetSettings();
				} catch (IOException e1) {
					e1.printStackTrace();
				}
								
			}});
		
		fileMenu.addSeparator();
		
		saveScreenShotMenuItem = new MenuItem("Save Screenshot...");
		fileMenu.add(saveScreenShotMenuItem);
		saveScreenShotMenuItem.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				
				saveScreenshot = saveScreenshotChooser.showSaveDialog(EmstApplication.this) == JFileChooser.APPROVE_OPTION;

			}
		});
		
		fileMenu.addSeparator();
		
		exitMenuItem = new MenuItem("Exit");
		fileMenu.add(exitMenuItem);
		exitMenuItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				dispose();				
			}});
				
		emstMenu = new Menu("EMST");
		menuBar.add(emstMenu);
		
		findEmstMenuItem = new MenuItem("Find EMST");
		emstMenu.add(findEmstMenuItem);
		findEmstMenuItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				if (graph != null)
					tree = Kruskal.mst(graph);
				else
					System.out.println("Cannot compute EMST until graph is loaded.");
				
			}});
		
		validateStMenuItem = new MenuItem("Validate ST");
		emstMenu.add(validateStMenuItem);
		validateStMenuItem.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				if (tree != null && graph != null)
					if (!VerifyST.check(tree, graph)) {
						JOptionPane.showMessageDialog(EmstApplication.this, "Not valid ST.");
					} else {
						JOptionPane.showMessageDialog(EmstApplication.this, "Found valid ST.");
					}
				else
					System.out.println("Warning: Cannot verify tree because either or both of the grahp and tree were not loaded.");
				
			}
		});
		
		viewMenu = new Menu("View");
		menuBar.add(viewMenu);
		
		graphViewMenu = new Menu("Graph");
		viewMenu.add(graphViewMenu);
		
		graphViewVertexColorMenuItem = new MenuItem("Choose Vertex Color...");
		graphViewMenu.add(graphViewVertexColorMenuItem);
		graphViewVertexColorMenuItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				graphVertexColor = JColorChooser.showDialog(EmstApplication.this, "Choose Vertex Color", graphVertexColor==null?Color.cyan:graphVertexColor);
				
			}});
		
		
		graphViewEdgeColorMenuItem = new MenuItem("Choose Edge Color...");
		graphViewMenu.add(graphViewEdgeColorMenuItem);
		graphViewEdgeColorMenuItem.addActionListener(new ActionListener() {
			
			@Override
			public void actionPerformed(ActionEvent e) {
				graphEdgeColor = JColorChooser.showDialog(EmstApplication.this, "Choose Edge Color", graphEdgeColor==null?Color.white:graphEdgeColor);
			}
		});
		
		graphViewMenu.addSeparator();
		
		graphViewToggleGraphMenuItem = new CheckboxMenuItem("Show Graph");
		graphViewMenu.add(graphViewToggleGraphMenuItem);
		graphViewToggleGraphMenuItem.setState(true);
		
		emstViewMenu = new Menu("EMST");
		viewMenu.add(emstViewMenu);
		
		emstViewColorMenuItem = new MenuItem("Choose Color...");
		emstViewMenu.add(emstViewColorMenuItem);
		emstViewColorMenuItem.addActionListener(new ActionListener(){

			@Override
			public void actionPerformed(ActionEvent e) {
				emstEdgeColor = JColorChooser.showDialog(EmstApplication.this, "Choose Tree Edge Color", emstEdgeColor==null?Color.red:emstEdgeColor);
				
			}});
		
		emstViewToggleEmst = new CheckboxMenuItem("Show EMST");
		emstViewMenu.add(emstViewToggleEmst);
		emstViewToggleEmst.setState(true);
		
		setMenuBar(menuBar);
	}
	

	private void resetSettings() {
		emstViewToggleEmst.setState(true);
		graphViewToggleGraphMenuItem.setState(true);
	}

	
	// WindowListener methods.
	
	@Override
	public void windowOpened(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowClosing(WindowEvent e) {
		dispose();
		
	}

	@Override
	public void windowClosed(WindowEvent e) {
		System.exit(0);		
	}

	@Override
	public void windowIconified(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowDeiconified(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowActivated(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void windowDeactivated(WindowEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void display(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();
		
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
		gl.glLoadIdentity();
		
		camera.setPerspective(gl);
		
		gl.glTranslatef(-1.5f, 0.0f, -6.0f);
		gl.glTranslatef(3.0f, 0.0f, 0.0f);
		
		if (graph != null) {			
			if (graphViewToggleGraphMenuItem.getState())
				graph.draw(gl, graphEdgeColor, graphVertexColor);
			else
				 graph.drawVertices(gl, graphVertexColor);
		}
		
		if (tree != null && emstViewToggleEmst.getState()) {
			gl.glBegin(GL.GL_LINES);
			gl.glColor3f(emstEdgeColor.getRed()/255f, emstEdgeColor.getGreen()/255f, emstEdgeColor.getBlue()/255f);
			
			for (Edge edge: tree) {
				Float v1[] = edge.getV1().getCoordinates();
				Float v2[] = edge.getV2().getCoordinates();
				if (v1.length == 3)
					gl.glVertex3f(v1[0], v1[1], v1[2]);
				else
					gl.glVertex3f(v1[0], v1[1], 0.0f);
				if (v2.length == 3)
					gl.glVertex3f(v2[0], v2[1], v2[2]);
				else 
					gl.glVertex3f(v2[0], v2[1], 0.0f);
			}
			gl.glEnd();
		}
		
		if (saveScreenshot) {
			saveScreenshot = false;
			try {
				Screenshot.writeToFile(saveScreenshotChooser.getSelectedFile(), glCanvas.getWidth(), glCanvas.getHeight());
			} catch(Throwable t) {
				t.printStackTrace();
			}
		}
		
	}

	@Override
	public void displayChanged(GLAutoDrawable arg0, boolean arg1, boolean arg2) {
		// TODO Auto-generated method stub
		System.out.println("displayChanged");
	}

	@Override
	public void init(GLAutoDrawable drawable) {
		GL gl = drawable.getGL();
		gl.glShadeModel(GL.GL_SMOOTH);

		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

		// Set up the depth buffer
		gl.glClearDepth(1.0f);
		gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glDepthFunc(GL.GL_LEQUAL);

		gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);

		drawable.addMouseListener(this);
		drawable.addMouseMotionListener(this);

		camera = new Camera(.10f, .10f, 0.0f, 0.0f, 0.0f, 5.0f);
		
	}

	@Override
	public void reshape(GLAutoDrawable drawable, int x, int y, int width,
			int height) {
	

		if (height == 0)
			height = 1;

		GL gl = drawable.getGL();

		gl.glViewport(0, 0, width, height);

		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		
		glu.gluPerspective(45.0f, (float) width / (float) height, 1f,
				100000.0f);

		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();
	
	}

	@Override
	public void keyTyped(KeyEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void keyPressed(KeyEvent e) {
		float velocity = 0.0f;
		if (e.getKeyChar()=='w')
			velocity = 0.05f;
		else if (e.getKeyChar()=='s')
			velocity = -0.05f;
		else if (e.getKeyChar() == 'S')
			velocity = -0.005f;
		else if (e.getKeyChar() == 'W')
			velocity = 0.005f;
		else if (e.getKeyChar() == 'q')
			velocity = 0.5f;
		else if (e.getKeyChar() == 'e')
			velocity = -0.5f;
		else return;
		camera.changeVelocity(velocity);
		//System.out.println("velocity: " + .5f);
		
	}

	@Override
	public void keyReleased(KeyEvent e) {
		camera.changeVelocity(0.0f);
		
	}

	@Override
	public void mouseDragged(MouseEvent e) {
		camera.changeHeading(.4f * (e.getX() - mouseClickX));
		mouseClickX = e.getX();

		camera.changePitch(.4f * (e.getY() - mouseClickY));
		mouseClickY = e.getY();
		
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseClicked(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mousePressed(MouseEvent e) {
		mouseClickX = e.getX();
		mouseClickY = e.getY();		
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}
	
}
