package emst.ds;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.media.opengl.GL;

import org.apache.log4j.Logger;

public class Graph {

	private static final Logger logger = Logger.getLogger(Graph.class);
	
	protected List<Vertex> vertices;
	protected List<Edge> edges;
	protected String name;
	
	public Graph(List<Vertex> vertices, List<Edge> edges, String name) {
		this.vertices = vertices;
		this.edges = edges;
		this.name = name;
		
		for (Edge e : this.edges) {
			e.computeWeight();
		}
	}
	
	public Graph() {
	}

	public List<Vertex> getVertices() {
		return vertices;
	}
	public void setVertices(List<Vertex> vertices) {
		this.vertices = vertices;
	}
	public List<Edge> getEdges() {
		return edges;
	}
	public void setEdges(List<Edge> edges) {
		this.edges = edges;
	}
	public String getName() {
		return name;
	}
	public void setName(String name) {
		this.name = name;
	}
	
	public static Graph loadFile(File file) throws IOException {
		FileReader freader = null;
		BufferedReader br = null;
		
		List<Vertex> vertices = new ArrayList<Vertex>();
		List<Edge> edges = new ArrayList<Edge>();
		
		Hashtable<String, Vertex> verticesDict = new Hashtable<String, Vertex>();
		
		try {
			freader = new FileReader(file);
			br = new BufferedReader(freader);
			
			String line = br.readLine();
			
			while (line != null && !line.equalsIgnoreCase("vertices {")) {
				line = br.readLine();
			}
			
			if (line == null) {
				throw new RuntimeException("Did not find \"vertices\" declaration.");
			}
			
			line = br.readLine();
			
			while (line != null && !line.equals("}")) {

				String[] values = line.split(" ");
				
				Float[] coordinates = new Float[values.length -1];
				
				for (int i = 1; i < values.length; i++) {
					coordinates[i-1] = Float.parseFloat(values[i]);
				}
				
				Vertex v = new Vertex(values[0], coordinates);
		
				vertices.add(v);
				
				verticesDict.put(v.getName(), v);
				
				line = br.readLine();
			}
			
			if (line == null)
				throw new RuntimeException("Did not find \"edges\" declaration.");
			
			while (line != null && !line.equalsIgnoreCase("edges {")) {
				line = br.readLine();
			}
			
			if (line == null)
				throw new RuntimeException("Did not find \"edges\" declaration.");
			
			line = br.readLine();
			
			
			while (line != null && !line.equals("}")) {
				String[] values = line.split(" ");
				
				Edge edge = new Edge(values[0], verticesDict.get(values[1]), verticesDict.get(values[2]));
				edges.add(edge);
				
				line = br.readLine();
			}
			
			
		} finally {
			if (freader != null) {
				freader.close();
			}
		}
		
		Graph pslg = new Graph(vertices, edges, file.getName());
		
		
		return pslg;
		
	}
	
	public void draw(GL gl, Color edgeColor, Color vertexColor) {
		gl.glBegin(GL.GL_LINES);
		gl.glColor3f(edgeColor.getRed()/255f, edgeColor.getGreen()/255f, edgeColor.getBlue()/255f);
		
		for (Edge edge : edges) {
			Float v1[] = edge.getV1().getCoordinates();
			Float v2[] = edge.getV2().getCoordinates();
			if (v1.length == 3)
				gl.glVertex3f(v1[0], v1[1], v1[2]);
			else
				gl.glVertex3f(v1[0], v1[1], 0.0f);
			if (v2.length == 3)
				gl.glVertex3f(v2[0], v2[1], v2[2]);
			else 
				gl.glVertex3f(v2[0], v2[1], 0.0f);
		}
		gl.glEnd();
		
		drawVertices(gl, vertexColor);
	}
	
	public void drawVertices(GL gl, Color vertexColor) {
		gl.glPointSize(4.0f);
		gl.glBegin(GL.GL_POINTS);
		gl.glColor3f(vertexColor.getRed()/255f, vertexColor.getGreen()/255f, vertexColor.getBlue()/255f);		
		
		for (Vertex vertex : vertices) {
			Float v[] = vertex.getCoordinates();
			gl.glVertex3f(v[0], v[1], (v.length==2?0.0f:v[2]));				
		}
		gl.glEnd();
		
	}
	
	public void removeDuplicateEdges() {
		Map<String, String> g = new HashMap<String, String>();
		
		Iterator<Edge> itr = edges.iterator();
		
		while (itr.hasNext()) {
			Edge e = itr.next();
			String key = e.getV1().getName() + e.getV2().getName();
			String key2 = e.getV2().getName() + e.getV1().getName();
			if (g.get(key) == null && g.get(key2) == null) {
				g.put(key, "1");				
				g.put(key2, "1");
			} else {
				itr.remove();
			}
		}
		
	}
	
	public void removeDuplicateVertices() {
		Map<String, String> vs = new HashMap<String, String>();
		
		Iterator<Vertex> itr = vertices.iterator();
		
		while (itr.hasNext()) {
			Vertex v = itr.next();
			StringBuilder key = new StringBuilder();
			
			for (Float i : v.getCoordinates()) {
				key.append(i).append(",");
			}
			
			if (vs.get(key.toString()) == null) {
				vs.put(key.toString(), "1");
			} else {
				itr.remove();
			}						
		}
	}
	
	public void removeOrphanedEdges() {
		Iterator<Edge> itr = edges.iterator();
				
		while(itr.hasNext()) {
			Edge edge = itr.next();
			
			if (findVertexName(edge.getV1().getName()) == false || findVertexName(edge.getV2().getName()) == false)
				itr.remove();
			
		}
	}
	
	public List<Edge> findIncidentEdges(Vertex v) {
		List<Edge> incident = new ArrayList<Edge>();
		
		for (Edge edge : edges) {
			if (edge.getV1().getName().equals(v.getName()) || edge.getV2().getName().equals(v.getName()))
				incident.add(edge);
		}
		
		if (logger.isDebugEnabled()) {
			StringBuilder sb = new StringBuilder();
			sb.append("All edges incident on ").append(v).append("\n");
			
			for (Edge edge : incident) {
				sb.append(edge.getName()).append("\t").append(edge.getV1().getName());
				
				List<Vertex> v1Duplicates = findDuplicateVertices(edge.getV1());
				List<Vertex> v2Duplicates = findDuplicateVertices(edge.getV2());
				
				if (v1Duplicates.size() > 0) {
					sb.append("(");
					for (Vertex vduplicate : v1Duplicates) {
						sb.append(vduplicate.getName());
						sb.append(",");
					}
					
					sb.deleteCharAt(sb.lastIndexOf(","));
					sb.append(")");
				}
				
				
				
				sb.append("\t").append(edge.getV2().getName());
				
				if (v2Duplicates.size() > 0) {
					sb.append("(");
					for (Vertex vduplicate : v2Duplicates) {
						sb.append(vduplicate.getName());
						sb.append(",");
					}
					
					sb.deleteCharAt(sb.lastIndexOf(","));
					sb.append(")");
				}
				sb.append("\n");				
			}
			
			logger.debug(sb.toString());
		}
		
		return incident;
	}
	
	public List<Vertex> findDuplicateVertices(Vertex v) {
		List<Vertex> duplicates = new ArrayList<Vertex>();
		
		for (Vertex candidate : vertices) {
			if (candidate.equals(v))
				continue;
			
			if (candidate.sameCoordinates(v))
				duplicates.add(candidate);
		}
		
		return duplicates;
	}
	
	public String formatDuplicateVerticesInGraph() {
		
		StringBuilder b = new StringBuilder();
		
		for (Vertex v : vertices) {
			List<Vertex> duplicates = findDuplicateVertices(v);
			
			b.append(v.getName()).append("\t");
			
			if (duplicates.size() > 0) {
				b.append("(");
				
				for (Vertex dupe : duplicates) {
					b.append(dupe.getName());
					b.append(",");					
				}
				b.deleteCharAt(b.lastIndexOf(","));
				b.append(")");
			}
			b.append("\n");
		}
		
		return b.toString();
	}
	
	public boolean findVertexName(String vertexName) {
		for (Vertex v : vertices) {
			if (v.getName().equalsIgnoreCase(vertexName))
				return true;
		}
		return false;
	}
	
	public boolean find(Vertex vertex) {
		for (Vertex v : vertices) {
			if (v.equals(vertex))
				return true;
		}
		
		return false;
	}

	public void mergeVertices(Vertex v1, Vertex v2) {
		// v2 will always be merged into v1
		
		List<Edge> edgesOnV2 = findIncidentEdges(v2);
		
		for (Edge edge : edgesOnV2) {
			
			Vertex ev1 = edge.getV1();
			Vertex ev2 = edge.getV2();
			
			if (ev1.equals(v2))
				edge.setV1(v1);
			else if (ev2.equals(v2))
				edge.setV2(v1);
			else
				throw new IllegalStateException("Should not have gotten here.");					
		}
		

	}
	
	public void mergeAllDuplicateVertices() {
		List<Vertex> verticesToBeRemoved = new ArrayList<Vertex>();
		for (Vertex v : vertices) {
			
			if (verticesToBeRemoved.contains(v))
				continue;
			
			List<Vertex> duplicates = findDuplicateVertices(v);
			
			Iterator<Vertex> itr = duplicates.iterator();
			while (itr.hasNext()) {
				Vertex dupe = itr.next();
				mergeVertices(v, dupe);
				verticesToBeRemoved.add(dupe);
			}
		}
				
		
		vertices.removeAll(verticesToBeRemoved);
	}
	
	public void cleanUpGraph() {
		mergeAllDuplicateVertices();
		removeDuplicateEdges();
	}
	
}
